/*
 * Copyright (C) 2000 Robert Fitzsimons
 */

#include "datatypes.h"
#include "debug.h"
#include "classfile.h"

#include <stdlib.h>  /* FIXME RF 2000-09-23.  I need to use a different method to allocate memory. */

ClassFile* classfile_build(kuint8* buffer, kuint buffer_length) {
	ClassFile* classfile = NULL;
	kuint8** buffer_pointer = &buffer;
	kuint index = 0;
	kuint index2 = 0;

	classfile = calloc(1, sizeof(ClassFile));

	classfile->magic = read_u4(buffer_pointer);
	if(classfile->magic != 0xCAFEBABE) {
		free(classfile);
		return NULL;
	}

	classfile->minor_version = read_u2(buffer_pointer);
	classfile->major_version = read_u2(buffer_pointer);

	classfile->constant_count = read_u2(buffer_pointer);
	if(classfile->constant_count > 0) {
		classfile->constant = calloc(classfile->constant_count, sizeof(ConstantInfo));
		for(index = 1; index < classfile->constant_count; index++) {
			ConstantInfo* constant = &classfile->constant[index];
			constant->info.tag = read_u1(buffer_pointer);
			switch(constant->info.tag) {
				case CONSTANT_Info_Utf8:
					constant->utf8_info.length = read_u2(buffer_pointer);
					constant->utf8_info.bytes = (*buffer_pointer);
					(*buffer_pointer) += constant->utf8_info.length;
					break;
				case CONSTANT_Info_Integer:
				case CONSTANT_Info_Float:
					constant->v32_info.bytes = read_u4(buffer_pointer);
					break;
				case CONSTANT_Info_Long:
				case CONSTANT_Info_Double:
					constant->v64_info.high_bytes = read_u4(buffer_pointer);
					constant->v64_info.low_bytes = read_u4(buffer_pointer);
					index++;
					break;
				case CONSTANT_Info_Class:
					constant->class_info.name_index = read_u2(buffer_pointer);
					break;
				case CONSTANT_Info_String:
					constant->string_info.string_index = read_u2(buffer_pointer);
					break;
				case CONSTANT_Info_FieldRef:
				case CONSTANT_Info_MethodRef:
				case CONSTANT_Info_InterfaceMethodRef:
					constant->ref_info.class_index = read_u2(buffer_pointer);
					constant->ref_info.nametype_index = read_u2(buffer_pointer);
					break;
				case CONSTANT_Info_NameAndType:
					constant->nametype_info.name_index = read_u2(buffer_pointer);
					constant->nametype_info.descriptor_index = read_u2(buffer_pointer);
					break;
			}
		}
	}

	classfile->access_flags = read_u2(buffer_pointer);
	classfile->this_class = read_u2(buffer_pointer);
	classfile->super_class = read_u2(buffer_pointer);

	classfile->interface_count = read_u2(buffer_pointer);
	if(classfile->interface_count > 0) {
		classfile->interface = calloc(classfile->interface_count, sizeof(kuint));
		for(index = 0; index < classfile->interface_count; index++) {
			classfile->interface[index] = read_u2(buffer_pointer);
		}
	}

	classfile->field_count = read_u2(buffer_pointer);
	if(classfile->field_count > 0) {
		classfile->field = calloc(classfile->field_count, sizeof(FieldInfo));
		for(index = 0; index < classfile->field_count; index++) {
			FieldInfo* field = &classfile->field[index];
			field->access_flags = read_u2(buffer_pointer);
			field->name_index = read_u2(buffer_pointer);
			field->descriptor_index = read_u2(buffer_pointer);
			field->attribute_count = read_u2(buffer_pointer);
			if(field->attribute_count > 0) {
				field->attribute = calloc(field->attribute_count, sizeof(AttributeInfo));
				for(index2 = 0; index2 < field->attribute_count; index2++) {
					AttributeInfo* attribute = &field->attribute[index2];
					attribute->name_index = read_u2(buffer_pointer);
					attribute->length = read_u4(buffer_pointer);
					attribute->array = (*buffer_pointer);
					(*buffer_pointer) += attribute->length;
				}
			}
		}
	}

	classfile->method_count = read_u2(buffer_pointer);
	if(classfile->method_count > 0) {
		classfile->method = calloc(classfile->method_count, sizeof(MethodInfo));
		for(index = 0; index < classfile->method_count; index++) {
			MethodInfo* method = &classfile->method[index];
			method->access_flags = read_u2(buffer_pointer);
			method->name_index = read_u2(buffer_pointer);
			method->descriptor_index = read_u2(buffer_pointer);
			method->attribute_count = read_u2(buffer_pointer);
			if(method->attribute_count > 0) {
				method->attribute = calloc(method->attribute_count, sizeof(AttributeInfo));
				for(index2 = 0; index2 < method->attribute_count; index2++) {
					AttributeInfo* attribute = &method->attribute[index2];
					attribute->name_index = read_u2(buffer_pointer);
					attribute->length = read_u4(buffer_pointer);
					attribute->array = (*buffer_pointer);
					(*buffer_pointer) += attribute->length;
				}
			}
		}
	}

	classfile->attribute_count = read_u2(buffer_pointer);
	if(classfile->attribute_count > 0) {
		classfile->attribute = calloc(classfile->attribute_count, sizeof(AttributeInfo));
		for(index2 = 0; index2 < classfile->attribute_count; index2++) {
			AttributeInfo* attribute = &classfile->attribute[index2];
			attribute->name_index = read_u2(buffer_pointer);
			attribute->length = read_u4(buffer_pointer);
			attribute->array = (*buffer_pointer);
			(*buffer_pointer) += attribute->length;
		}
	}

	return classfile;
}

void classfile_free(ClassFile* classfile) {
	kuint index = 0;

	if(classfile->constant_count > 0) {
		free(classfile->constant);
	}

	if(classfile->interface_count > 0) {
		free(classfile->interface);
	}

	if(classfile->field_count > 0) {
		for(index = 0; index < classfile->field_count; index++) {
			if(classfile->field[index].attribute_count > 0) {
				free(classfile->field[index].attribute);
			}
		}
		free(classfile->field);
	}

	if(classfile->method_count > 0) {
		for(index = 0; index < classfile->method_count; index++) {
			if(classfile->method[index].attribute_count > 0) {
				free(classfile->method[index].attribute);
			}
		}
		free(classfile->method);
	}

	if(classfile->attribute_count > 0) {
		free(classfile->attribute);
	}

	free(classfile);
}

void classfile_print(ClassFile* classfile) {
	kuint index = 0;
	kuint index2 = 0;

	debug(
		"ClassFile\n%08X\n%08X\n%08X\n",
		classfile->magic,
		classfile->minor_version,
		classfile->major_version
	);

	debug("%08X\n", classfile->constant_count);
	for(index = 1; index < classfile->constant_count; index++) {
		ConstantInfo* constant = &classfile->constant[index];
		switch(constant->info.tag) {
			case CONSTANT_Info_Utf8:
				debug(
					"  %08X  %08X  %08X  %08X\n",
					index,
					constant->info.tag,
					constant->utf8_info.length,
					constant->utf8_info.bytes
				);
				break;
			case CONSTANT_Info_Integer:
			case CONSTANT_Info_Float:
				debug(
					"  %08X  %08X  %08X\n",
					index,
					constant->info.tag,
					constant->v32_info.bytes
				);
				break;
			case CONSTANT_Info_Long:
			case CONSTANT_Info_Double:
				debug(
					"  %08X  %08X  %08X  %08X\n",
					index,
					constant->info.tag,
					constant->v64_info.high_bytes,
					constant->v64_info.low_bytes
				);
				index++;
				break;
			case CONSTANT_Info_Class:
				debug(
					"  %08X  %08X  %08X\n",
					index,
					constant->info.tag,
					constant->class_info.name_index
				);
				break;
			case CONSTANT_Info_String:
				debug(
					"  %08X  %08X  %08X\n",
					index,
					constant->info.tag,
					constant->string_info.string_index
				);
				break;
			case CONSTANT_Info_FieldRef:
			case CONSTANT_Info_MethodRef:
			case CONSTANT_Info_InterfaceMethodRef:
				debug(
					"  %08X  %08X  %08X  %08X\n",
					index,
					constant->info.tag,
					constant->ref_info.class_index,
					constant->ref_info.nametype_index
				);
				break;
			case CONSTANT_Info_NameAndType:
				debug(
					"  %08X  %08X  %08X  %08X\n",
					index,
					constant->info.tag,
					constant->nametype_info.name_index,
					constant->nametype_info.descriptor_index
				);
				break;
		}
	}

	debug(
		"%08X\n%08X\n%08X\n",
		classfile->access_flags,
		classfile->this_class,
		classfile->super_class
	);

	debug("%08X\n", classfile->interface_count);
	for(index = 0; index < classfile->interface_count; index++) {
		debug("  %08X\n", classfile->interface[index]);
	}

	debug("%08X\n", classfile->field_count);
	for(index = 0; index < classfile->field_count; index++) {
		FieldInfo* field = &classfile->field[index];
		debug(
			"  %08X  %08X  %08X  %08X  %08X\n",
			index,
			field->access_flags,
			field->name_index,
			field->descriptor_index,
			field->attribute_count
		);
		for(index2 = 0; index2 < field->attribute_count; index2++) {
			AttributeInfo* attribute = &field->attribute[index2];
			debug(
				"    %08X  %08X  %08X  %08X\n",
				index2,
				attribute->name_index,
				attribute->length,
				attribute->array
			);
		}
	}

	debug("%08X\n", classfile->method_count);
	for(index = 0; index < classfile->method_count; index++) {
		MethodInfo* method = &classfile->method[index];
		debug(
			"  %08X  %08X  %08X  %08X  %08X\n",
			index,
			method->access_flags,
			method->name_index,
			method->descriptor_index,
			method->attribute_count
		);
		for(index2 = 0; index2 < method->attribute_count; index2++) {
			AttributeInfo* attribute = &method->attribute[index2];
			debug(
				"    %08X  %08X  %08X  %08X\n",
				index2,
				attribute->name_index,
				attribute->length,
				attribute->array
			);
		}
	}

	debug("%08X\n", classfile->attribute_count);
	for(index2 = 0; index2 < classfile->attribute_count; index2++) {
		AttributeInfo* attribute = &classfile->attribute[index2];
		debug(
			"  %08X  %08X  %08X  %08X\n",
			index2,
			attribute->name_index,
			attribute->length,
			attribute->array
		);
	}
}

kuint read_u1(kuint8** buffer_pointer) {
	return (*(*buffer_pointer)++);
}

kuint read_u2(kuint8** buffer_pointer) {
	kuint value = 0;
	value |= ((*(*buffer_pointer)++) << 8);
	value |= (*(*buffer_pointer)++);
	return value;
}

kuint read_u4(kuint8** buffer_pointer) {
	kuint value = 0;
	value |= ((*(*buffer_pointer)++) << 24);
	value |= ((*(*buffer_pointer)++) << 16);
	value |= ((*(*buffer_pointer)++) << 8);
	value |= (*(*buffer_pointer)++);
	return value;
}

